@@ -1,7 +1,7 @@
 package Seq;
 use 5.036;
-our $VERSION = '0.003';
-use subs 'bind', 'join', 'select', 'last';
+our $VERSION = '0.005';
+use subs 'bind', 'join', 'select', 'last', 'sort';
 use Scalar::Util qw(reftype);
 use List::Util;
 use Carp qw(croak);
@@ -12,11 +12,12 @@ use Sub::Exporter -setup => {
 };
 use DDP;
 
-# TODO: contains?, firstIndex?, mapX, reduce?, sort, cache
+# TODO: contains?, firstIndex?, mapX, sort, cache
 #       interspers, slice, zip, unzip, foldBack, any,
 #       forall, none, average, average_by,
 #       pairwise, windowed, transpose, item, chunk_by_size,
 #       one, minmax, minmax_by,
+#       regex_match, regex_replace
 #
 # + ways to do regexes on strings
 # + using this module to scan file-system
@@ -28,9 +29,9 @@ use DDP;
 # o good way to also implement async with it?
 
 # Important functions used in FP code. So adding them.
-sub id ($x)     { return $x          }
-sub fst($array) { return $array->[0] }
-sub snd($array) { return $array->[1] }
+sub id  :prototype($) { return $_[0]    }
+sub fst :prototype($) { return $_[0][0] }
+sub snd :prototype($) { return $_[0][1] }
 
 # a helper to create a function for selecting a hash-key
 sub key :prototype($) {
@@ -48,7 +49,18 @@ sub assign :prototype(&) {
 
 # creates a sequence from a subroutine
 sub from_sub($class, $f) {
-    return bless($f, 'Seq');
+    return bless(sub {
+        my $abort = 0;
+        my $it    = $f->();
+        return sub {
+            return undef if $abort;
+            if ( defined(my $x = $it->()) ) {
+                return $x;
+            }
+            $abort = 1;
+            undef $it;
+        }
+    }, 'Seq');
 }
 
 # always an empty sequence
@@ -60,7 +72,7 @@ sub empty($class) {
 
 # TODO: When $state is a reference. Same handling as in fold?
 sub unfold($class, $state, $f) {
-    return bless(sub {
+    from_sub('Seq', sub {
         # IMPORTANT: Perl signatures are aliases. As we assign
         # to $state later, we need to make a copy here.
         # Removing this lines causes bugs.
@@ -74,7 +86,7 @@ sub unfold($class, $state, $f) {
             $abort = 1 if not defined $x;
             return $x;
         }
-    }, 'Seq');
+    });
 }
 
 sub init($class, $count, $f) {
@@ -132,8 +144,10 @@ sub from_array($class, $xs) {
 # Pass it a hashref. The function $f is executed for every
 # (key, value) pair. The result is used as a single item in
 # the sequence.
+#
+# Seq->from_hash : Hash<'Key, 'Value> -> ('Key -> 'Value -> 'a) -> Seq<'a>
 sub from_hash($class, $hashref, $f) {
-    return bless(sub {
+    from_sub('Seq', sub {
         my $idx  = 0;
         my @keys = keys %$hashref;
         my $last = $#keys;
@@ -142,7 +156,7 @@ sub from_hash($class, $hashref, $f) {
             my $key = $keys[$idx++];
             return $f->($key, $hashref->{$key});
         }
-    }, 'Seq');
+    });
 }
 
 # Concatenates a list of Seq into a single Seq
@@ -160,11 +174,11 @@ sub concat($class, @seqs) {
 #- Methods
 #    functions operating on Seq and returning another Seq
 
-sub append($iterA, $iterB) {
-    return bless(sub {
+sub append($seqA, $seqB) {
+    from_sub('Seq', sub {
         my $exhaustedA = 0;
-        my $itA = $iterA->();
-        my $itB = $iterB->();
+        my $itA = $seqA->();
+        my $itB = $seqB->();
 
         return sub {
             REDO:
@@ -181,26 +195,26 @@ sub append($iterA, $iterB) {
                 }
             }
         };
-    }, 'Seq');
+    });
 }
 
 # map : Seq<'a> -> ('a -> 'b) -> Seq<'b'>
-sub map($iter, $f) {
-    return bless(sub {
-        my $it = $iter->();
+sub map($seq, $f) {
+    from_sub('Seq', sub {
+        my $it = $seq->();
         return sub {
             if ( defined(my $x = $it->()) ) {
                 return $f->($x);
             }
             return undef;
         }
-    }, 'Seq');
+    });
 }
 
 # bind : Seq<'a> -> ('a -> Seq<'b>) -> Seq<'b>
-sub bind($iter, $f) {
-    return bless(sub {
-        my $it   = $iter->();
+sub bind($seq, $f) {
+    from_sub('Seq', sub {
+        my $it   = $seq->();
         my $seqB = undef;
 
         return sub {
@@ -221,7 +235,7 @@ sub bind($iter, $f) {
             # when $seqB is not defined and $it does not return new values
             return undef;
         }
-    }, 'Seq');
+    });
 }
 
 # flatten : Seq<Seq<'a>> -> Seq<'a>
@@ -229,6 +243,13 @@ sub flatten($iter) {
     return bind($iter, \&id);
 }
 
+# flatten_array : Seq<Array<'a>> -> Seq<'a>
+sub flatten_array($seq) {
+    return bind($seq, sub($array) {
+        return from_array('Seq', $array);
+    });
+}
+
 # cartesian : Seq<'a> -> Seq<'b> -> Seq<'a * 'b>
 sub cartesian($seqA, $seqB) {
     bind($seqA, sub($a) {
@@ -253,9 +274,7 @@ sub join($seqA, $seqB, $predicate) {
 # merge : Seq<'a * 'b> -> ('a -> 'b -> 'c) -> Seq<'c>
 sub merge($iter, $merge) {
     bind($iter, sub($tuple) {
-        my ($a, $b) = @$tuple;
-        my $c = $merge->($a, $b);
-        return wrap(Seq => $c);
+        return wrap(Seq => $merge->($tuple->[0], $tuple->[1]));
     });
 }
 
@@ -347,7 +366,7 @@ sub select($iter, $mapA, $mapB) {
 
 # combines map and filter
 sub choose($iter, $chooser) {
-    return bless(sub {
+    from_sub('Seq', sub {
         my $it = $iter->();
         return sub {
             while ( defined(my $x = $it->()) ) {
@@ -356,7 +375,7 @@ sub choose($iter, $chooser) {
             }
             return undef;
         }
-    }, 'Seq');
+    });
 }
 
 sub mapi($iter, $f) {
@@ -364,7 +383,7 @@ sub mapi($iter, $f) {
 }
 
 sub filter($iter, $predicate) {
-    return bless(sub {
+    from_sub('Seq', sub {
         my $it = $iter->();
         return sub {
             while ( defined(my $x = $it->()) ) {
@@ -372,11 +391,11 @@ sub filter($iter, $predicate) {
             }
             return undef;
         }
-    }, 'Seq');
+    });
 }
 
 sub take($iter, $amount) {
-    return bless(sub {
+    from_sub('Seq', sub {
         my $i             = $iter->();
         my $returnedSoFar = 0;
         return sub {
@@ -388,11 +407,11 @@ sub take($iter, $amount) {
             }
             return;
         }
-    }, 'Seq');
+    });
 }
 
 sub skip($iter, $amount) {
-    return bless(sub {
+    from_sub('Seq', sub {
         my $it = $iter->();
         my $count = 0;
         return sub {
@@ -401,7 +420,7 @@ sub skip($iter, $amount) {
             }
             return $it->();
         }
-    }, 'Seq');
+    });
 }
 
 sub indexed($iter) {
@@ -412,7 +431,7 @@ sub indexed($iter) {
 }
 
 sub distinct_by($iter, $f) {
-    return bless(sub {
+    from_sub('Seq', sub {
         my $it = $iter->();
         my %seen;
         return sub {
@@ -425,7 +444,7 @@ sub distinct_by($iter, $f) {
             }
             return undef;
         }
-    }, 'Seq');
+    });
 }
 
 # remove duplicates - it uses a hash to remember seen items
@@ -435,6 +454,34 @@ sub distinct($iter) {
     return distinct_by($iter, \&id);
 }
 
+# TODO: Instead of fsts and snds provide a function to pick the index of an array.
+#       Maybe even a function to pick and re-order multiple elements
+#         Like: ->pick([3,1,5])
+
+sub fsts($seq) {
+    return $seq->map(sub ($x) { $x->[0] });
+}
+
+sub snds($seq) {
+    return $seq->map(sub ($x) { $x->[1] });
+}
+
+# TODO: zip can handle a list of sequences
+sub zip($seqA, $seqB) {
+    from_sub('Seq', sub {
+        my $itA = $seqA->();
+        my $itB = $seqB->();
+
+        return sub {
+            if (defined (my $a = $itA->())) {
+            if (defined (my $b = $itB->())) {
+                    return [$a,$b];
+            }}
+            return undef;
+        }
+    });
+}
+
 #- Side-Effects
 #    functions that have side-effects or produce side-effects. Those are
 #    immediately executed, usually consuming all elements of Seq at once.
@@ -461,7 +508,7 @@ sub do($iter, $f) {
 }
 
 sub rev($iter) {
-    return bless(sub {
+    from_sub('Seq', sub {
         my @list = to_list($iter);
         return sub {
             if ( defined(my $x = pop @list) ) {
@@ -469,27 +516,88 @@ sub rev($iter) {
             }
             return undef;
         };
-    }, 'Seq');
+    });
+}
+
+sub sort($seq, $comparer) {
+    from_sub('Seq', sub {
+        local ($a, $b);
+        my @array = CORE::sort { $comparer->($a, $b) } to_list($seq);
+        my $idx   = 0;
+        my $count = @array;
+
+        return sub {
+            if ( $idx < $count ) {
+                my $x = $array[$idx];
+                $idx++;
+                return $x;
+            }
+            return undef;
+        };
+    });
+}
+
+# group_by : Seq<'a> -> ('a -> 'Key) -> Seq<Seq<'a>>
+sub group_by($seq, $get_key) {
+    my %group;
+    iter($seq, sub($a) {
+        my $key = $get_key->($a);
+        push $group{$key}->@*, $a;
+    });
+
+    return from_hash('Seq', \%group, sub($key, $value) {
+        return from_array('Seq', $value);
+    });
+}
+
+# group_fold :
+#   Seq<'a>
+#   -> (unit -> 'State)
+#   -> ('a -> 'Key)
+#   -> ('State -> 'a -> 'State)
+#   -> Seq<'State>
+sub group_fold($seq, $get_state, $get_key, $folder) {
+    my %group;
+    iter($seq, sub($a) {
+        my $key = $get_key->($a);
+        push $group{$key}->@*, $a;
+    });
+
+    return from_hash('Seq', \%group, sub($key, $array) {
+        my $state = $get_state->();
+        for my $a ( @$array ) {
+            $state = $folder->($state, $a);
+        }
+        return $state;
+    });
 }
 
 #- Converter
 #    Those are functions converting Seq to none Seq types
 
-sub fold($iter, $state, $folder) {
-    # when $state is reference, we assume $folder mutate $state
-    if ( ref $state ) {
-        iter($iter, sub($x) {
-            $folder->($state, $x);
-        });
-        return $state;
-    }
-    # otherwise $folder returns new $state
-    else {
-        iter($iter, sub($x) {
-            $state = $folder->($state, $x);
-        });
-        return $state;
-    }
+# fold is like a foreach-loop. You iterate through all items generating
+# a new 'State. The $folder is passed the latest
+# 'State and one 'a from the sequence. You return
+# the next 'State that should be used. Once all elements of 'a
+# sequence are processed. The last 'State is returned.
+#
+# fold : Seq<'a> -> 'State -> ('State -> 'a -> 'State) -> 'State
+sub fold($seq, $state, $folder) {
+    iter($seq, sub($x) {
+        $state = $folder->($state, $x);
+    });
+    return $state;
+}
+
+# Same as fold. But when you mutate 'State and return 'State from
+# the lambda. You can use this function instead.
+#
+# fold : Seq<'a> -> 'State -> ('State -> 'a -> 'State) -> 'State
+sub fold_mut($seq, $state, $folder) {
+    iter($seq, sub($x) {
+        $folder->($state, $x);
+    });
+    return $state;
 }
 
 # Like fold, but without an initial state. When sequence is empty
@@ -497,20 +605,20 @@ sub fold($iter, $state, $folder) {
 # left to right to produce an output. Needs a sequence
 # with one item to work properly. It is encouraged to use `fold`
 # instead.
-sub reduce($seq, $reducer) {
-    my $f = first(take($seq, 1));
-
-    return fold(skip($seq, 1), $f, $reducer) if defined $f;
-    return undef;
+# reduce: Seq<'a> -> ('a -> 'a -> 'a) -> 'a
+sub reduce($seq, $reducer, $default) {
+    return fold(skip($seq, 1), first($seq, $default), $reducer);
 }
 
-sub first($seq, $default=undef) {
+# first : Seq<'a> -> 'a -> 'a
+sub first($seq, $default) {
     my $first = $seq->()();
     return $first if defined $first;
     return $default;
 }
 
-sub last($seq, $default=undef) {
+# last : Seq<'a> -> 'a -> 'a
+sub last($seq, $default) {
     my $last;
     iter($seq, sub($x) {
         $last = $x;
@@ -519,44 +627,48 @@ sub last($seq, $default=undef) {
     return $default;
 }
 
-sub to_array($iter) {
+# to_array : Seq<'a> -> Array<'a>
+sub to_array($seq) {
     state $folder = sub($array, $x) { push @$array, $x };
-    return fold($iter, [], $folder);
+    return fold_mut($seq, [], $folder);
 }
 
-sub to_list($iter) {
-    return @{ to_array($iter) };
+sub to_list($seq) {
+    return @{ to_array($seq) };
 }
 
-sub count($iter) {
+# count : Seq<'a> -> int
+sub count($seq) {
     state $folder = sub($count, $x) { $count+1 };
-    return fold($iter, 0, $folder);
+    return fold($seq, 0, $folder);
 }
 
-sub sum($iter) {
+# sum : Seq<'a> -> float
+sub sum($seq) {
     state $folder = sub($sum, $x) { $sum + $x };
-    return fold($iter, 0, $folder);
+    return fold($seq, 0, $folder);
 }
 
-sub sum_by($iter, $f) {
-    return fold($iter, 0, sub($sum, $x) {
+# sum_by : Seq<'a> -> (float -> 'a -> float) -> float
+sub sum_by($seq, $f) {
+    return fold($seq, 0, sub($sum, $x) {
         return $sum + $f->($x);
     });
 }
 
 # returns the min value or undef on empty sequence
 # min value is compared with numerical <
-sub min($seq, $default=undef) {
+sub min($seq, $default) {
     min_by($seq, \&id, $default);
 }
 
 # returns the min value or undef on empty sequence
 # min value is compared using lt
-sub min_str($seq, $default=undef) {
+sub min_str($seq, $default) {
     min_by_str($seq, \&id, $default);
 }
 
-sub min_by($seq, $key, $default=undef) {
+sub min_by($seq, $key, $default) {
     return fold($seq, undef, sub($min, $x) {
         my $value = $key->($x);
         defined $min
@@ -565,7 +677,7 @@ sub min_by($seq, $key, $default=undef) {
     }) // $default;
 }
 
-sub min_by_str($seq, $key, $default=undef) {
+sub min_by_str($seq, $key, $default) {
     return fold($seq, undef, sub($min, $x) {
         my $value = $key->($x);
         defined $min
@@ -575,15 +687,15 @@ sub min_by_str($seq, $key, $default=undef) {
 }
 
 # returns the max value or undef when sequence is empty
-sub max($seq, $default=undef) {
+sub max($seq, $default) {
     max_by($seq, \&id, $default);
 }
 
-sub max_str($seq, $default=undef) {
+sub max_str($seq, $default) {
     max_by_str($seq, \&id, $default);
 }
 
-sub max_by($seq, $key, $default=undef) {
+sub max_by($seq, $key, $default) {
     return fold($seq, undef, sub($max, $x) {
         my $value = $key->($x);
         defined $max
@@ -592,7 +704,7 @@ sub max_by($seq, $key, $default=undef) {
     }) // $default;
 }
 
-sub max_by_str($seq, $key, $default=undef) {
+sub max_by_str($seq, $key, $default) {
     return fold($seq, undef, sub($max, $x) {
         my $value = $key->($x);
         defined $max
@@ -607,9 +719,9 @@ sub str_join($iter, $sep) {
 
 # Build a hash by providing a keying function. Later elements
 # in the sequence overwrite previous one.
-sub to_hash($iter, $get_key) {
+sub to_hash($seq, $get_key) {
     my %hash;
-    iter($iter, sub($x) {
+    iter($seq, sub($x) {
         $hash{$get_key->($x)} = $x;
     });
     return \%hash;
@@ -617,21 +729,34 @@ sub to_hash($iter, $get_key) {
 
 # Build a hash by providing a keying function. Values
 # are put into arrays to allow key with multiple values.
-sub group_by($iter, $get_key) {
+#
+# to_hash_of_array: Seq<'a> -> ('a -> 'Key) -> Hash<'Key, Array<'a>>
+sub to_hash_of_array($seq, $get_key) {
     my %hash;
-    iter($iter, sub($x) {
+    iter($seq, sub($x) {
         push @{ $hash{$get_key->($x)} }, $x;
     });
     return \%hash;
 }
 
+# to_array_of_array : Seq<Seq<'a>> -> Array<Array<'a>>
+sub to_array_of_array($seq) {
+    my @outer;
+    iter($seq, sub($inner) {
+        push @outer, to_array($inner);
+    });
+    return \@outer;
+}
+
 # returns first element for which the given $predicate returns true
-sub find($iter, $predicate) {
-    my $it = $iter->();
+#
+# find : Seq<'a> -> ('a -> bool) -> 'a
+sub find($seq, $default, $predicate) {
+    my $it = $seq->();
     while ( defined(my $x = $it->()) ) {
         return $x if $predicate->($x);
     }
-    return;
+    return $default;
 }
 
 1;
